import logging
import inspect

from django.db.models.signals import post_delete, m2m_changed
from django.dispatch import receiver
from django.apps import apps
from django.db import models
from functools import wraps
from typing import Any, Optional, List

from .middleware import get_current_user
from .settings import UNREGISTERED_CLASSES

logger = logging.getLogger("audit.crud")

EVENT_TYPES = [
    "CREATE",
    "UPDATE",
    "DELETE",
    "BULK_CREATE",
    "BULK_UPDATE",
    "M2M",
]


def should_audit(instance):
    """Return True or False to indicate whether the instance should be audited."""
    # do not audit any model listed in UNREGISTERED_CLASSES
    for unregistered_class in UNREGISTERED_CLASSES:
        if isinstance(instance, unregistered_class):
            return False
    return True


def get_user_details():
    user = get_current_user()
    data = {
        "id": str(user.id) if hasattr(user, "id") else "",
        "title": user.title if hasattr(user, "title") else "",
        "email": user.email if hasattr(user, "email") else "",
        "first_name": user.first_name if hasattr(user, "first_name") else "",
        "middle_name": user.middle_name if hasattr(user, "middle_name") else "",
        "last_name": user.last_name if hasattr(user, "last_name") else "",
        "sex": user.sex if hasattr(user, "sex") else "",
        "date_of_birth": user.date_of_birth if hasattr(user, "date_of_birth") else "",
    }
    return data


def get_calling_model() -> Optional[str]:
    """Get the model name from the calling function's frame."""
    try:
        # Get the current frame
        frame = inspect.currentframe()
        # Go up 3 frames to get to the actual calling function
        # (1 for get_calling_model, 1 for the signal handler, 1 for the bulk operation)
        for _ in range(3):
            frame = frame.f_back
            if frame is None:
                return None

        # Get the calling function's name
        calling_function = frame.f_code.co_name
        # Get the module name
        module_name = frame.f_globals.get("__name__", "")

        # Check if this is a direct bulk operation call
        if "bulk_create" in calling_function or "bulk_update" in calling_function:
            return module_name.split(".")[-1]
    except Exception:
        pass
    return None


class ModelSignalMixin:
    """Mixin to add signal handling capabilities to models."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self._original_m2m = self._get_m2m_state()

    def _get_m2m_state(self) -> dict:
        """Get the current state of M2M fields."""
        return {
            field.name: set(
                getattr(self, field.name).all().values_list("id", flat=True)
            )
            for field in self._meta.many_to_many
        }


def push_log(
    message: str,
    model: str,
    event_type: str,
    instance_id: str,
    extra: dict = {},
) -> None:
    payload: dict = {
        "model": model,
        "instance_id": str(instance_id),
        "event_type": event_type,
        "user": get_user_details(),
        "extra": extra,
    }
    logger.audit(message, extra=payload)


def patch_model_event(model_class: type[models.Model]) -> None:
    """Monkey patch a model to add signal handling capabilities."""

    if not issubclass(model_class, ModelSignalMixin):
        # Add the mixin to the model's base classes
        model_class.__bases__ = (ModelSignalMixin,) + model_class.__bases__

        # Store the original methods
        original_save = model_class.save
        original_bulk_create = models.QuerySet.bulk_create
        original_bulk_update = models.QuerySet.bulk_update

        @wraps(original_save)
        def save_with_signals(self: models.Model, *args: Any, **kwargs: Any) -> None:
            is_new = self._state.adding

            # Call the original save method
            original_save(self, *args, **kwargs)

            # Log the event
            event_type = EVENT_TYPES[0] if is_new else EVENT_TYPES[1]

            push_log(
                f"{event_type} event for {model_class.__name__} (id: {self.pk})",
                model_class.__name__,
                event_type,
                str(self.pk),
                {},
            )

        @wraps(original_bulk_create)
        def bulk_create_with_signals(
            self, objs: List[models.Model], *args: Any, **kwargs: Any
        ) -> List[models.Model]:
            if not objs:
                return original_bulk_create(self, objs, *args, **kwargs)

            # Get the calling model
            calling_model = get_calling_model()
            if not calling_model:
                return original_bulk_create(self, objs, *args, **kwargs)

            # Call the original bulk_create method
            created_objs = original_bulk_create(self, objs, *args, **kwargs)

            # Log only if this is the calling model
            if calling_model == model_class.__name__:
                first_obj = created_objs[0]
                push_log(
                    f"{EVENT_TYPES[3]} event for {model_class.__name__} (id: {first_obj.pk})",
                    model_class.__name__,
                    EVENT_TYPES[3],
                    str(first_obj.pk),
                    {
                        "total_count": len(created_objs),
                    },
                )

            return created_objs

        @wraps(original_bulk_update)
        def bulk_update_with_signals(
            self, objs: List[models.Model], fields: List[str], batch_size=None
        ) -> None:
            if not objs:
                return original_bulk_update(self, objs, fields, batch_size)

            # Call the original bulk_update method
            original_bulk_update(self, objs, fields, batch_size)

            # Get the calling model
            calling_model = get_calling_model()
            if not calling_model:
                return original_bulk_update(self, objs, fields, batch_size)

            # Log only if this is the calling model
            if calling_model == model_class.__name__:
                first_obj = objs[0]
                push_log(
                    f"{EVENT_TYPES[4]} event for {model_class.__name__}",
                    model_class.__name__,
                    EVENT_TYPES[4],
                    str(first_obj.pk),
                    {
                        "total_count": len(objs),
                        "fields": fields,
                    },
                )

        # Replace the methods
        model_class.save = save_with_signals
        models.QuerySet.bulk_create = bulk_create_with_signals
        models.QuerySet.bulk_update = bulk_update_with_signals

        # Add delete signal handling
        @receiver(post_delete, sender=model_class)
        def handle_delete(
            sender: type[models.Model], instance: models.Model, **kwargs: Any
        ) -> None:
            push_log(
                f"{EVENT_TYPES[2]} event for {model_class.__name__} (id: {instance.pk})",
                model_class.__name__,
                EVENT_TYPES[2],
                str(instance.pk),
            )

        # Add M2M signal handling
        for field in model_class._meta.many_to_many:

            @receiver(m2m_changed, sender=getattr(model_class, field.name).through)
            def handle_m2m_changed(
                sender: type[models.Model],
                instance: models.Model,
                action: str,
                pk_set: set,
                **kwargs: Any,
            ) -> None:
                if action not in ["post_add", "post_remove", "post_clear"]:
                    return

                field_name = kwargs.get("model", sender).__name__.lower()
                push_log(
                    f"M2M {action} event for {model_class.__name__} (id: {instance.pk})",
                    model_class.__name__,
                    EVENT_TYPES[5],
                    str(instance.pk),
                    {
                        "field_name": field_name,
                        "related_ids": list(pk_set) if pk_set else None,
                    },
                )


def setup_model_signals() -> None:
    """Set up signals for all models in the project."""
    for app_config in apps.get_app_configs():
        for model in app_config.get_models():
            if not should_audit(model):
                continue

            if not issubclass(model, ModelSignalMixin):
                patch_model_event(model)


# Initialize signals when the module is imported
setup_model_signals()



from .constants import CONSOLE_FORMAT
from .logger_levels import API, AUDIT, LOGIN


def get_console_formatter() -> dict:
    return {
        "format": CONSOLE_FORMAT,
    }


def get_json_formatter() -> dict:
    return {
        "()": "django_activity_audit.formatters.JsonFormatter",
    }


def get_api_formatter() -> dict:
    return {
        "()": "django_activity_audit.formatters.APIFormatter",
    }


def get_audit_formatter() -> dict:
    return {
        "()": "django_activity_audit.formatters.AuditFormatter",
    }


def get_login_formatter() -> dict:
    return {
        "()": "django_activity_audit.formatters.LoginFormatter",
    }


def get_json_handler(
    level: str,
    filename: str = "audit_logs/app.log",
    formatter: str = "json",
    max_bytes: int = 1024 * 1024 * 10,
    backup_count: int = 5,
) -> dict:
    return {
        "level": level,
        "class": "logging.handlers.RotatingFileHandler",
        "formatter": formatter,
        "filename": filename,
        "maxBytes": max_bytes,
        "backupCount": backup_count,
    }


def get_api_handler(
    filename: str = "audit_logs/api.log",
    formatter: str = "api_json",
) -> dict:
    return {
        "level": API,
        "class": "django_activity_audit.handlers.AuditLogHandler",
        "filename": filename,
        "maxBytes": 1024 * 1024 * 10,  # 10MB
        "backupCount": 5,
        "formatter": formatter,
    }


def get_audit_handler(
    filename: str = "audit_logs/audit.log",
    formatter: str = "audit_json",
) -> dict:
    return {
        "level": AUDIT,
        "class": "django_activity_audit.handlers.AuditLogHandler",
        "filename": filename,
        "maxBytes": 1024 * 1024 * 10,  # 10MB
        "backupCount": 5,
        "formatter": formatter,
    }


def get_login_handler(
    filename: str = "audit_logs/login.log",
    formatter: str = "login_json",
) -> dict:
    return {
        "level": LOGIN,
        "class": "django_activity_audit.handlers.AuditLogHandler",
        "filename": filename,
        "maxBytes": 1024 * 1024 * 10,  # 10MB
        "backupCount": 5,
        "formatter": formatter,
    }


def push_usage_log(message: str, event: str, success: bool, error: str, extra: dict):
    """
    data:
        - message: message
        - user: user details
        - event: login or logout
        - success: true or false
        - error: error message
        - extra: {
            - cognito_id: cognito id
            - status_code: status code
        }
    """
    import logging
    from .signals import get_user_details

    logger = logging.getLogger("audit.login")

    data = {
        "user": get_user_details(),
        "event": event,
        "success": success,
        "error": error,
        "extra": extra,
    }

    logger.login(
        message,
        extra=data,
    )



import logging

AUDIT = 21
API = 22
LOGIN = 23

# Only add the level if it doesn't already exist
if not hasattr(logging, "AUDIT"):
    logging.addLevelName(AUDIT, "AUDIT")

    # Add the audit method to Logger class
    def audit(self, message, *args, **kwargs):
        if self.isEnabledFor(AUDIT):
            self._log(AUDIT, message, args, **kwargs)

    logging.Logger.audit = audit
else:
    # Use existing AUDIT level if it's already defined
    AUDIT = logging.AUDIT

if not hasattr(logging, "API"):
    logging.addLevelName(API, "API")

    def api(self, message, *args, **kwargs):
        if self.isEnabledFor(API):
            self._log(API, message, args, **kwargs)

    logging.Logger.api = api
else:
    API = logging.API

if not hasattr(logging, "LOGIN"):
    logging.addLevelName(LOGIN, "LOGIN")

    def login(self, message, *args, **kwargs):
        if self.isEnabledFor(LOGIN):
            self._log(LOGIN, message, args, **kwargs)

    logging.Logger.login = login
else:
    LOGIN = logging.LOGIN



from django.conf import settings
from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType
from django.contrib.sessions.models import Session
from django.db.migrations import Migration
from django.db.migrations.recorder import MigrationRecorder
from django.apps import apps

UNREGISTERED_CLASSES = [
    Migration,
    Session,
    Permission,
    ContentType,
    MigrationRecorder.Migration,
]

# Import and unregister LogEntry class only if Django Admin app is installed
if apps.is_installed("django.contrib.admin"):
    from django.contrib.admin.models import LogEntry

    UNREGISTERED_CLASSES += [LogEntry]

# URL patterns to exclude from logging
UNREGISTERED_URLS = [r"^/admin/", r"^/static/", r"^/favicon.ico$"]
UNREGISTERED_URLS = getattr(
    settings, "django_activity_audit_UNREGISTERED_URLS_DEFAULT", UNREGISTERED_URLS
)
UNREGISTERED_URLS.extend(
    getattr(settings, "django_activity_audit_UNREGISTERED_URLS_EXTRA", [])
)

# URL patterns to include in logging (if empty, all URLs are logged)
REGISTERED_URLS = getattr(settings, "django_activity_audit_REGISTERED_URLS", [])



import time
import paramiko
import logging

from typing import Tuple, Optional
from requests.sessions import Session

from .constants import REQUEST_TYPES

logger = logging.getLogger("audit.request")

PROTOCOLS = ("http", "sftp")
OPERATIONS = ("upload", "download")
THIRTY_SECONDS_TIMEOUT = 30

"""
    log structure:
    {
        "timestamp": "2021-01-01 12:00:00.000",
        "level": "INFO",
        "name": "audit.request",
        "request_type": "external",
        "service_name": "default",
        "protocol": "http",
        "request_repr": {
            "endpoint": "https://example.com/api/v1/users",
            "method": "GET",
            "headers": {"Content-Type": "application/json"},
            "body": {"name": "John Doe", "email": "john.doe@example.com"},
        },
        "response_repr": {
            "status_code": 200,
            "body": {"name": "John Doe", "email": "john.doe@example.com"},
        },
        "error_message": "",
        "execution_time": 0,
    }
"""


class HTTPClient(Session):
    def __init__(self, service_name: str = "default"):
        super().__init__()
        self.log_payload = {
            "service_name": service_name,
            "protocol": PROTOCOLS[0],
            "request_type": REQUEST_TYPES[1],
            "request_repr": {},
            "response_repr": {},
            "error_message": "",
            "execution_time": 0,
        }

    def request(self, method, url, **kwargs):
        start_time = time.time()
        headers = kwargs.get("headers", {})
        data = kwargs.get("data", {})
        error = None
        response = None

        request_repr = {
            "endpoint": url,
            "method": method,
            "headers": headers,
            "body": data,
        }

        try:
            response = super().request(method, url, **kwargs)
            response_repr = {
                "status_code": response.status_code,
                "body": response.json(),
            }
        except Exception as e:
            error = str(e)

        execution_time = time.time() - start_time

        self.log_payload["request_repr"] = str(request_repr)
        self.log_payload["execution_time"] = execution_time

        if error:
            self.log_payload["error_message"] = error
        else:
            self.log_payload["response_repr"] = str(response_repr)

        logger.api("Audit External Service", extra=self.log_payload)

        return response


class SFTPClient:
    def __init__(
        self,
        host,
        port,
        username,
        password,
        service_name,
    ):
        self.client = paramiko.SSHClient()
        self.client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        self.host = host
        self.port = port
        self.username = username
        self.password = password
        self.service_name = service_name
        self.channel = None
        self.log_payload = {
            "service_name": service_name,
            "request_type": REQUEST_TYPES[1],
            "protocol": PROTOCOLS[1],
            "request_repr": {
                "host": host,
                "operation": None,
                "remote_path": None,
                "filename": None,
            },
            "response_repr": None,
            "error_message": None,
            "execution_time": 0,
        }

    def __create_log(self):
        # Add external service log
        pass

    def connect(
        self,
    ) -> Tuple[Optional[paramiko.SFTPClient], Optional[Exception]]:
        try:
            if not self.channel:
                self.client.connect(
                    self.host,
                    self.port,
                    self.username,
                    self.password,
                    timeout=THIRTY_SECONDS_TIMEOUT,
                )
                self.channel = self.client.open_sftp()
                logger.info("SFTP connection established successfully")
            logger.info("Reusing existing SFTP connection")
        except Exception as e:
            error_message = f"SFTP connection failed. Error: {str(e)}"
            logger.info(error_message)
            self.log_payload["error_message"] = error_message
            self.__create_log()
            return None, e

        return self.channel, None

    def is_valid_path(self, path_to_folder) -> Tuple[bool, Optional[str]]:
        if self.channel:
            try:
                self.channel.listdir(path_to_folder)
                logger.info(f"{path_to_folder} validated successfully")
                return True, None
            except Exception as _:
                return False, f"Folder({path_to_folder}) not found."
        return False, f"Connection not established"

    def upload(
        self, path_to_folder, filename, file_content
    ) -> Tuple[bool, Optional[str]]:
        start_time = time.time()
        result = ""

        self.log_payload["request_repr"]["operation"] = OPERATIONS[0]
        self.log_payload["request_repr"]["remote_path"] = path_to_folder
        self.log_payload["request_repr"]["filename"] = filename

        if self.channel:
            _, error = self.is_valid_path(path_to_folder)

            if not error:
                try:
                    with self.channel.open(
                        f"{path_to_folder}{filename}", "wb"
                    ) as remote_file:
                        remote_file.write(file_content)
                    result = f"{filename} uploaded successfully to {path_to_folder}"
                    logger.info(f"{result}")
                except Exception as e:
                    self.log_payload["error_message"] = (
                        f"File upload failed. Error: {str(e)}"
                    )
            self.log_payload["error_message"] = (
                f"Path validation failed. Error: {str(error)}"
            )
        else:
            self.log_payload["error_message"] = "Connection not established"

        execution_time = time.time() - start_time

        self.log_payload["response_repr"] = {"message": result}
        self.log_payload["execution_time"] = execution_time

        self.__create_log()

        return result

    def close(self):
        if self.channel:
            self.channel.close()
        self.client.close()



from django.apps import AppConfig


class AuditLoggingConfig(AppConfig):
    default_auto_field = "django.db.models.BigAutoField"
    name = "django_activity_audit"
    verbose_name = "Django Activity Audit"

    def ready(self):
        # Import and register custom log levels
        from . import logger_levels

        # Force registration of custom levels
        logger_levels.AUDIT
        logger_levels.API
        logger_levels.LOGIN

        # Initialize signals
        from . import signals  # noqa



import logging

from logging.handlers import RotatingFileHandler

from .formatters import APIFormatter, AuditFormatter, LoginFormatter
from .logger_levels import API, AUDIT, LOGIN


class AuditLogHandler(RotatingFileHandler):
    """Custom handler for audit.request logger that ensures proper formatting and validation of audit logs."""

    def __init__(
        self,
        filename,
        mode="a",
        maxBytes=0,
        backupCount=0,
        encoding=None,
        delay=False,
    ):
        super().__init__(filename, mode, maxBytes, backupCount, encoding, delay)
        if self.level == API:
            self.setFormatter(APIFormatter())
        elif self.level == AUDIT:
            self.setFormatter(AuditFormatter())
        elif self.level == LOGIN:
            self.setFormatter(LoginFormatter())

    def emit(self, record):
        """
        Emit a record with additional values for audit-specific fields.
        """
        try:
            # Handle extra if present
            if hasattr(record, "extra"):
                for key, value in record.extra.items():
                    setattr(record, key, value)

            super().emit(record)

        except Exception as e:
            self.handleError(record)
            # Log the error to the root logger
            logging.getLogger().error(f"Error in AuditLogHandler: {str(e)}")



default_app_config = "django_activity_audit.apps.AuditLoggingConfig"

from django_activity_audit.utils import (
    get_console_formatter,
    get_json_formatter,
    get_json_handler,
    get_api_formatter,
    get_api_handler,
    get_audit_formatter,
    get_audit_handler,
    get_login_formatter,
    get_login_handler,
)

__all__ = [
    "get_console_formatter",
    "get_json_formatter",
    "get_json_handler",
    "get_api_formatter",
    "get_api_handler",
    "get_audit_formatter",
    "get_audit_handler",
    "get_login_formatter",
    "get_login_handler",
]



import logging
import datetime
import json


class JsonFormatter(logging.Formatter):
    def __init__(self, timestamp_format: str = "%Y-%m-%d %H:%M:%S.%f"):
        super().__init__()
        self.timestamp_format = timestamp_format

    def format(
        self,
        record,
    ):
        log_data = {
            "timestamp": datetime.datetime.fromtimestamp(record.created).strftime(
                self.timestamp_format
            )[:-3],
            "level": record.levelname,
            "name": record.name,
            "path": record.pathname,
            "module": record.module,
            "function": record.funcName,
            "message": record.getMessage(),
            "exception": "",
            "request": "",
            "extra": "",
        }

        # Add exception info if present for ERROR
        if record.exc_info:
            log_data["exception"] = "{}".format(self.formatException(record.exc_info))

        # Add request info if available
        if hasattr(record, "request"):
            log_data["request"] = {
                "method": getattr(record.request, "method", None),
                "path": getattr(record.request, "path", None),
                "user": str(getattr(record.request, "user", None)),
            }

        # Add extra fields if present
        if hasattr(record, "extra"):
            log_data.update(record.extra)

        return json.dumps(log_data)


class APIFormatter(logging.Formatter):
    """Custom formatter for audit logs that ensures consistent JSON formatting."""

    def __init__(self, timestamp_format: str = "%Y-%m-%d %H:%M:%S.%f"):
        super().__init__()
        self.timestamp_format = timestamp_format

    def format(self, record):
        # Start with basic log data
        log_data = {
            "timestamp": datetime.datetime.fromtimestamp(record.created).strftime(
                self.timestamp_format
            )[:-3],
            "level": record.levelname,
            "name": record.name,
            "message": record.getMessage(),
        }

        # Add all audit-specific fields if they exist
        audit_fields = [
            "service_name",
            "request_type",
            "protocol",
            "request_repr",
            "response_repr",
            "error_message",
            "execution_time",
        ]

        for field in audit_fields:
            log_data[field] = getattr(record, field)

        return json.dumps(log_data)


class AuditFormatter(logging.Formatter):
    def __init__(self, timestamp_format: str = "%Y-%m-%d %H:%M:%S.%f"):
        super().__init__()
        self.timestamp_format = timestamp_format

    def format(self, record):
        log_data = {
            "timestamp": datetime.datetime.fromtimestamp(record.created).strftime(
                self.timestamp_format
            )[:-3],
            "level": record.levelname,
            "name": record.name,
            "message": record.getMessage(),
        }

        audit_fields = [
            "model",
            "event_type",
            "instance_id",
            "user",
            "extra",
        ]

        for field in audit_fields:
            log_data[field] = getattr(record, field)

        return json.dumps(log_data)


class LoginFormatter(logging.Formatter):
    def __init__(self, timestamp_format: str = "%Y-%m-%d %H:%M:%S.%f"):
        super().__init__()
        self.timestamp_format = timestamp_format

    def format(self, record):
        log_data = {
            "timestamp": datetime.datetime.fromtimestamp(record.created).strftime(
                self.timestamp_format
            )[:-3],
            "level": record.levelname,
            "name": record.name,
            "message": record.getMessage(),
        }

        audit_fields = [
            "user",
            "event",
            "success",
            "error",
            "extra",
        ]

        for field in audit_fields:
            log_data[field] = getattr(record, field)

        return json.dumps(log_data)



import time
import logging
import json
import re
import contextlib

from asgiref.local import Local
from asgiref.sync import iscoroutinefunction, markcoroutinefunction
from django.http import HttpResponse
from django.utils.deprecation import MiddlewareMixin

from .settings import UNREGISTERED_URLS, REGISTERED_URLS
from .constants import REQUEST_TYPES

logger = logging.getLogger("audit.request")

_thread_locals = Local()


class MockRequest:
    def __init__(self, *args, **kwargs):
        user = kwargs.pop("user", None)
        self.user = user
        super().__init__(*args, **kwargs)


def get_current_request():
    return getattr(_thread_locals, "request", None)


def set_current_request(request):
    _thread_locals.request = request


def get_current_user():
    request = get_current_request()
    if request:
        return getattr(request, "user", None)
    return None


def set_current_user(user):
    try:
        _thread_locals.request.user = user
    except AttributeError:
        request = MockRequest(user=user)
        _thread_locals.request = request


def clear_request():
    with contextlib.suppress(AttributeError):
        del _thread_locals.request


def should_log_url(url):
    # check if current url is blacklisted
    for unregistered_url in UNREGISTERED_URLS:
        pattern = re.compile(unregistered_url)
        if pattern.match(url):
            return False

    # only audit URLs listed in REGISTERED_URLS (if it's set)
    if len(REGISTERED_URLS) > 0:
        for registered_url in REGISTERED_URLS:
            pattern = re.compile(registered_url)
            if pattern.match(url):
                return True
        return False

    # all good
    return True


class AuditLoggingMiddleware(MiddlewareMixin):
    """
    log structure:
    {
        "timestamp": "2021-01-01 12:00:00.000",
        "level": "INFO",
        "name": "audit.request",
        "service_name": "default",
        "protocol": "http",
        "request_repr": {
            "endpoint": "https://example.com/api/v1/users",
            "method": "GET",
            "headers": {"Content-Type": "application/json"},
            "body": {"name": "John Doe", "email": "john.doe@example.com"},
        },
        "response_repr": {
            "status_code": 200,
            "body": {"name": "John Doe", "email": "john.doe@example.com"},
        },
        "error_message": "",
        "execution_time": 0,
    }
    """

    def __init__(self, get_response):
        self.get_response = get_response
        self.log_data = {
            "service_name": "review_board",
            "request_type": REQUEST_TYPES[0],
            "protocol": None,
            "request_repr": {},
            "response_repr": {},
            "error_message": None,
            "execution_time": 0,
        }

        if iscoroutinefunction(self.get_response):
            markcoroutinefunction(self)

    def __call__(self, request):
        if iscoroutinefunction(self):
            return self.__acall__(request)
        set_current_request(request)

        if not should_log_url(request.path):
            return self.get_response(request)

        start_time = time.time()

        # Log request
        request_data = {
            "method": request.method,
            "path": request.path,
            "query_params": dict(request.GET.items()),
            "headers": dict(request.headers),
            "user": str(request.user) if request.user.is_authenticated else None,
        }

        if request.content_type == "application/json":
            try:
                body = json.loads(request.body)
                request_data["body"] = body
            except json.JSONDecodeError:
                request_data["body"] = "Invalid JSON"

        # Get response
        response = self.get_response(request)
        end_time = time.time()

        # TODO: Find way to add status code to response_data

        # Log response
        response_data = {
            "headers": dict(response.headers),
        }

        if isinstance(response, HttpResponse):
            try:
                content = response.content.decode("utf-8")
                if content:
                    try:
                        response_data["body"] = json.loads(content)
                    except json.JSONDecodeError:
                        response_data["body"] = content
            except UnicodeDecodeError:
                response_data["body"] = "Binary content"

        self.log_data["execution_time"] = end_time - start_time
        self.log_data["protocol"] = "https" if request.is_secure() else "http"
        self.log_data["request_repr"] = request_data
        self.log_data["response_repr"] = response_data

        logger.api("Audit Internal Request", extra=self.log_data)

        clear_request()

        return response

    async def __acall__(self, request):
        set_current_request(request)

        if not should_log_url(request.path):
            return await self.get_response(request)

        start_time = time.time()

        # Log request
        request_data = {
            "method": request.method,
            "path": request.path,
            "query_params": dict(request.GET.items()),
            "headers": dict(request.headers),
            "user": str(request.user) if request.user.is_authenticated else None,
        }

        if request.content_type == "application/json":
            try:
                body = json.loads(request.body)
                request_data["body"] = body
            except json.JSONDecodeError:
                request_data["body"] = "Invalid JSON"

        # Get response
        response = await self.get_response(request)
        end_time = time.time()

        # TODO: Find way to add status code to response_data

        # Log response
        response_data = {
            "headers": dict(response.headers),
        }

        if isinstance(response, HttpResponse):
            try:
                content = response.content.decode("utf-8")
                if content:
                    try:
                        response_data["body"] = json.loads(content)
                    except json.JSONDecodeError:
                        response_data["body"] = content
            except UnicodeDecodeError:
                response_data["body"] = "Binary content"

        self.log_data["execution_time"] = end_time - start_time
        self.log_data["protocol"] = "https" if request.is_secure() else "http"
        self.log_data["request_repr"] = request_data
        self.log_data["response_repr"] = response_data

        logger.api("Audit Internal Request", extra=self.log_data)

        clear_request()

        return response
